import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env
import mujoco_py

class SwimmerEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self, file_path, seed=0,):
        mujoco_env.MujocoEnv.__init__(self, file_path, 8)
        utils.EzPickle.__init__(self)

    def step(self, a):
        ctrl_cost_coeff = 0.0001
        xposbefore = self.sim.data.qpos[0]
        self.do_simulation(a, self.frame_skip)
        xposafter = self.sim.data.qpos[0]
        reward_fwd = (xposafter - xposbefore) / self.dt
        reward_ctrl = -ctrl_cost_coeff * np.square(a).sum()
        reward = reward_fwd + reward_ctrl
        ob = self._get_obs()
        return ob, reward, False, dict(reward_fwd=reward_fwd, reward_ctrl=reward_ctrl)

    def _get_obs(self):
        qpos = self.sim.data.qpos
        qvel = self.sim.data.qvel
        # print('env',np.concatenate([qpos.flat[2:], qvel.flat]).shape)
        return np.concatenate([qpos.flat[2:], qvel.flat])

    def reset_model(self):
        self.set_state(
            self.init_qpos
            + self.np_random.uniform(low=-0.1, high=0.1, size=self.model.nq),
            self.init_qvel
            + self.np_random.uniform(low=-0.1, high=0.1, size=self.model.nv),
        )
        return self._get_obs()
    @property
    def physics(self):
        # check mujoco version is greater than version 1.50 to call correct physics
        # model containing PyMjData object for getting and setting position/velocity
        # check https://github.com/openai/mujoco-py/issues/80 for updates to api

        if mujoco_py.get_version() >= '1.50':
          return self.sim
        else:
          return self.model

    # def get_ori(self):
    #     ori = [0, 1, 0, 0]
    #     rot = self.physics.data.qpos[self.__class__.ORI_IND:self.__class__.ORI_IND + 4]  # take the quaternion
    #     ori = q_mult(q_mult(rot, ori), q_inv(rot))[1:3]  # project onto x-y plane
    #     ori = math.atan2(ori[1], ori[0])
    #     return ori

    def set_xy(self, xy):
        qpos = np.copy(self.physics.data.qpos)
        qpos[0] = xy[0]
        qpos[1] = xy[1]

        qvel = self.physics.data.qvel
        self.set_state(qpos, qvel)

    def get_xy(self):
        return self.physics.data.qpos[:2]